"""
Created on 2021-04-27.

@author: Mike K
"""
from hdf5tools import H5
import os
import pytest
from glob import glob
import xarray as xr

##############################################
### Parameters

base_path = os.path.join(os.path.split(os.path.realpath(os.path.dirname(__file__)))[0], 'datasets')


######################################
### Testing

files = glob(base_path + '/*.nc')
files.sort()

ds_ids = set([os.path.split(f)[-1].split('_')[0] for f in files])

# for ds_id in ds_ids:
#     ds_files = [xr.open_dataset(f, engine='h5netcdf') for f in files if ds_id in f]
#     h1 = H5(ds_files)
#     print(h1)
#     new_path = os.path.join(base_path, ds_id + '_test1.h5')
#     h1.to_hdf5(new_path)
#     x1 = xr.open_dataset(new_path, engine='h5netcdf')
#     print(x1)

#     first_times = x1.time.values[0:5]
#     x1.close()
#     h2 = h1.sel({'time': slice(first_times[0], first_times[-1])})
#     print(h2)
#     h2.to_hdf5(new_path)
#     x1 = xr.open_dataset(new_path, engine='h5netcdf')
#     print(x1.load())
#     assert x1.time.shape[0] == 4

#     main_vars = [v for v in list(x1.data_vars) if set(x1[v].dims) == set(x1.dims)]
#     x1.close()
#     h2 = h1.sel(include_data_vars=main_vars)
#     print(h2)
#     h2.to_hdf5(new_path)
#     x1 = xr.open_dataset(new_path, engine='h5netcdf')
#     print(x1.load())
#     assert set(x1.data_vars) == set(main_vars)
#     x1.close()

#     os.remove(new_path)

# for ds_id in ds_ids:
#     ds_files = []

#     for i, f in enumerate(files):
#         if ds_id in f:
#             if i == 0:
#                 ds_files.append(f)
#             else:
#                 ds_files.append(xr.open_dataset(f, engine='h5netcdf'))

#     h1 = H5(ds_files)
#     print(h1)
#     new_path = os.path.join(base_path, ds_id + '_test1.h5')
#     h1.to_hdf5(new_path)
#     x1 = xr.open_dataset(new_path, engine='h5netcdf')
#     print(x1)

#     first_times = x1.time.values[0:5]
#     x1.close()
#     h2 = h1.sel({'time': slice(first_times[0], first_times[-1])})
#     print(h2)
#     h2.to_hdf5(new_path)
#     x1 = xr.open_dataset(new_path, engine='h5netcdf')
#     print(x1.load())
#     assert x1.time.shape[0] == 4

#     main_vars = [v for v in list(x1.data_vars) if set(x1[v].dims) == set(x1.dims)]
#     x1.close()
#     h2 = h1.sel(include_data_vars=main_vars)
#     print(h2)
#     h2.to_hdf5(new_path)
#     x1 = xr.open_dataset(new_path, engine='h5netcdf')
#     print(x1.load())
#     assert set(x1.data_vars) == set(main_vars)
#     x1.close()

#     os.remove(new_path)


@pytest.mark.parametrize('ds_id', ds_ids)
def test_H5_xr(ds_id):
    """

    """
    ds_files = [xr.open_dataset(f, engine='h5netcdf') for f in files if ds_id in f]
    h1 = H5(ds_files)
    print(h1)
    new_path = os.path.join(base_path, ds_id + '_test1.h5')
    h1.to_hdf5(new_path)
    x1 = xr.open_dataset(new_path, engine='h5netcdf')
    print(x1)

    first_times = x1.time.values[0:5]
    x1.close()
    h2 = h1.sel({'time': slice(first_times[0], first_times[-1])})
    print(h2)
    h2.to_hdf5(new_path)
    x1 = xr.open_dataset(new_path, engine='h5netcdf')
    print(x1.load())
    assert x1.time.shape[0] == 4

    main_vars = [v for v in list(x1.data_vars) if set(x1[v].dims) == set(x1.dims)]
    x1.close()
    h2 = h1.sel(include_data_vars=main_vars)
    print(h2)
    h2.to_hdf5(new_path)
    x1 = xr.open_dataset(new_path, engine='h5netcdf')
    print(x1.load())
    assert set(x1.data_vars) == set(main_vars)
    x1.close()

    os.remove(new_path)


@pytest.mark.parametrize('ds_id', ds_ids)
def test_H5_hdf5(ds_id):
    """

    """
    ds_files = [f for f in files if ds_id in f]
    h1 = H5(ds_files)
    print(h1)
    new_path = os.path.join(base_path, ds_id + '_test1.h5')
    h1.to_hdf5(new_path)
    x1 = xr.open_dataset(new_path, engine='h5netcdf')
    print(x1)

    first_times = x1.time.values[0:5]
    x1.close()
    h2 = h1.sel({'time': slice(first_times[0], first_times[-1])})
    print(h2)
    h2.to_hdf5(new_path)
    x1 = xr.open_dataset(new_path, engine='h5netcdf')
    print(x1.load())
    assert x1.time.shape[0] == 4

    main_vars = [v for v in list(x1.data_vars) if set(x1[v].dims) == set(x1.dims)]
    x1.close()
    h2 = h1.sel(include_data_vars=main_vars)
    print(h2)
    h2.to_hdf5(new_path)
    x1 = xr.open_dataset(new_path, engine='h5netcdf')
    print(x1.load())
    assert set(x1.data_vars) == set(main_vars)
    x1.close()

    os.remove(new_path)


@pytest.mark.parametrize('ds_id', ds_ids)
def test_H5_mix(ds_id):
    """

    """
    ds_files = []

    for i, f in enumerate(files):
        if ds_id in f:
            if i == 0:
                ds_files.append(f)
            else:
                ds_files.append(xr.open_dataset(f, engine='h5netcdf'))

    h1 = H5(ds_files)
    print(h1)
    new_path = os.path.join(base_path, ds_id + '_test1.h5')
    h1.to_hdf5(new_path)
    x1 = xr.open_dataset(new_path, engine='h5netcdf')
    print(x1)

    first_times = x1.time.values[0:5]
    x1.close()
    h2 = h1.sel({'time': slice(first_times[0], first_times[-1])})
    print(h2)
    h2.to_hdf5(new_path)
    x1 = xr.open_dataset(new_path, engine='h5netcdf')
    print(x1.load())
    assert x1.time.shape[0] == 4

    main_vars = [v for v in list(x1.data_vars) if set(x1[v].dims) == set(x1.dims)]
    x1.close()
    h2 = h1.sel(include_data_vars=main_vars)
    print(h2)
    h2.to_hdf5(new_path)
    x1 = xr.open_dataset(new_path, engine='h5netcdf')
    print(x1.load())
    assert set(x1.data_vars) == set(main_vars)
    x1.close()

    os.remove(new_path)











# -*- coding: utf-8 -*-
"""
Utility functions.
"""
import os
import io
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point, Polygon, box, LineString
import shapely
from scipy.spatial import KDTree
import zstandard as zstd
import pathlib
from copy import copy
import smart_open
import smart_open.http as so_http
from time import sleep
import concurrent.futures

so_http.DEFAULT_BUFFER_SIZE = 524288

#########################################
### parameters

file_dict = {'btree.node_coords.np.zstd': 'https://b2.tethys-ts.xyz/file/nz-rec/btree.node_coords.np.zstd',
             'catch.blt': 'https://b2.tethys-ts.xyz/file/nz-rec/catch.blt',
             'node.blt': 'https://b2.tethys-ts.xyz/file/nz-rec/node.blt',
             'node_way_index.blt': 'https://b2.tethys-ts.xyz/file/nz-rec/node_way_index.blt',
             'way.blt': 'https://b2.tethys-ts.xyz/file/nz-rec/way.blt',
             'way_index.blt': 'https://b2.tethys-ts.xyz/file/nz-rec/way_index.blt',
             'way_tag.blt': 'https://b2.tethys-ts.xyz/file/nz-rec/way_tag.blt',
             'btree.node_ids.np.zstd': 'https://b2.tethys-ts.xyz/file/nz-rec/btree.node_ids.np.zstd'
             }


########################################
### Functions


def url_to_file(url, file_path, chunk_size: int=524288, retries=3):
    """
    General function to get an object from an S3 bucket. One of s3, connection_config, or public_url must be used.

    Parameters
    ----------
    url: http str
        The http url to the file.
    chunk_size: int
        The amount of bytes to download as once.

    Returns
    -------
    file object
        file object of the S3 object.
    """
    transport_params = {'buffer_size': chunk_size, 'timeout': 120}

    ## Get the object
    counter = retries
    while True:
        try:
            file_obj = smart_open.open(url, 'rb', transport_params=transport_params, compression='disable')
            file_path1 = pathlib.Path(file_path)
            file_path1.parent.mkdir(parents=True, exist_ok=True)

            with open(file_path1, 'wb') as f:
                chunk = file_obj.read(chunk_size)
                while chunk:
                    f.write(chunk)
                    chunk = file_obj.read(chunk_size)
            break
        except Exception as err:
            counter = counter - 1
            if counter > 0:
                sleep(3)
            else:
                print('smart_open could not open url with the following error:')
                print(err)
                file_obj = None
                break

    return file_obj


def check_files(data_path):
    """

    """
    data_path1 = pathlib.Path(data_path)

    files = [f.name for f in data_path1.glob('*') if f.name in file_dict]
    missing_files = []
    for f in file_dict:
        if f not in files:
            missing_files.append(file_dict[f])

    return missing_files


def download_files(data_path, only_missing=True):
    """

    """
    if only_missing:
        urls = check_files(data_path)
    else:
        urls = list(file_dict.values())

    print('Downloading: {}'.format(', '.join([os.path.split(url)[-1] for url in urls])))
    with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
        futures = []
        for url in urls:
            file_name = os.path.split(url)[-1]
            new_path = os.path.join(data_path, file_name)
            f = executor.submit(url_to_file, url, new_path)
            futures.append(f)
        _ = concurrent.futures.wait(futures)


def get_down_way(node_id, node_way, way):
    """

    """
    base_ways0 = set(node_way[int(node_id)])

    for way_id in base_ways0:
        nodes = way[int(way_id)]
        if node_id != nodes[-1]:
            way_down = way_id
            break

    return way_down


def make_kdtree(coords):
    """

    """
    kdtree = KDTree(coords)

    return kdtree


def nearest(kdtree, coords, nodes, node_way, way, max_distance=np.inf):
    """
    Function to find the nearest nodes and ways from coords. Uses the scipy function cKDTree.

    Parameters
    ----------
    kdtree : KDTree init object
        Already initialized KDTree.
    coords : np.array
        An array of (2) shape for coordiantes to query.
    nodes : np.array
        An array of node ids ordered by the original data of the kdtree.
    node_way : dict-like
        A dict of nodes to ways.
    max_distance : non-negative int, optional
        Return only neighbors within this distance. This is used to prune tree searches, so if you are doing a series of nearest-neighbor queries, it may help to supply the distance to the nearest neighbor of the most recent point.

    Returns
    -------
    distance, nodes, ways
    """
    coords1 = (np.asarray(coords) * 10000000).round().astype('int32')

    if isinstance(max_distance, int):
    	max_distance1 = 100 * max_distance
    else:
        max_distance1 = np.inf

    dist, idx = kdtree.query(coords1, k=1, distance_upper_bound=max_distance1)
    out_node = nodes[idx]
    out_way = get_down_way(out_node, node_way, way)

    return int(round(dist*0.01, 0)), out_node, out_way


def nearest_within_catchments(kdtree, coords, nodes, node, node_way, way, catch, max_distance=np.inf):
    """

    """
    coords1 = (np.asarray(coords) * 10000000).round().astype('int32')

    if isinstance(max_distance, int):
    	max_distance1 = 100 * max_distance
    else:
        max_distance1 = np.inf

    k = [1]
    while True:
        dist, idx = kdtree.query(coords1, k=k, distance_upper_bound=max_distance1)
        out_node = nodes[idx[0]]
        out_way = get_down_way(out_node, node_way, way)
        out_catch = catch[out_way]
        check = out_catch.intersects(Point(node[out_node] * 0.0000001))
        if check:
            distance = int(round(dist[0]*0.01, 0))
            break
        else:
            if k[0] > 100:
                raise ValueError('Cannot find any catchments.')
            k = [k[0] + 1]

    return distance, out_node, out_way


def find_upstream(node_id, node_way, way, way_index):
    """

    """
    ways_up = set()

    base_ways0 = set(node_way[int(node_id)])

    base_ways = set()
    no_ways = set()
    for way_id in base_ways0:
        nodes = way[int(way_id)]
        if node_id != nodes[0]:
            base_ways.add(way_id)
        else:
            no_ways.add(way_id)

    ways_up.update(base_ways)

    while base_ways:
        base_ways1 = copy(base_ways)
        base_ways = set()
        for way_id in base_ways1:
            new_ways = set(way_index[int(way_id)])
            new_ways = new_ways.difference(no_ways)
            base_ways.update(new_ways.difference(ways_up))
            ways_up.update(new_ways)

    return ways_up


def find_downstream(node_id, node_way, way, way_index):
    """

    """
    way_down = get_down_way(node_id, node_way, way)

    ways_down = [way_down]
    append = ways_down.append

    while True:
        down_node = way[int(way_down)][-1]

        down_ways = list(node_way[int(down_node)])
        down_ways.remove(way_down)

        if down_ways:
            for down_way in down_ways:
                up_node = way[int(down_way)][0]
                if up_node == down_node:
                    append(down_way)
                    way_down = down_way
                    break
        else:
            break

    return ways_down


# def make_reaches(ways, way, node, output='shapely'):
#     """

#     """
#     way_ids = []
#     reach_list = []
#     for way_id in ways:
#         nodes = way[int(way_id)]
#         reach = np.array([node[int(node_id)] for node_id in nodes])

#         if output in ('shapely', 'geopandas'):
#             reach = LineString(reach)

#         reach_list.append(reach)
#         way_ids.append(way_id)

#     if output == 'geopandas':
#         reach_out = gpd.GeoDataFrame(way_ids, geometry=reach_list, columns=['way_id'], crs=4326)
#     else:
#         reach_out = dict(zip(way_ids, reach_list))

#     return reach_out


# def make_catchment(ways_up, catch):
#     """

#     """
#     catch_list = []
#     append = catch_list.append
#     for way_id in ways_up:
#         catch0 = catch[int(way_id)]
#         append(catch0)

#     catch1 = shapely.ops.unary_union(catch_list).buffer(0.0001)

#     return catch1





# -*- coding: utf-8 -*-
"""
Functions to delineate catchments.
"""
import os
import numpy as np
import pandas as pd
from shapely.geometry import Point, Polygon, box, LineString
import shapely
import geopandas as gpd
from scipy.spatial import KDTree
import zstandard as zstd
import orjson
import pathlib
import booklet

# import utils
from . import utils

#####################################################
### Parameters


#####################################################
#### MFE REC streams network


class Water:
    """

    """
    def __init__(self, data_path, download_files=False, only_missing=True):
        """

        """
        ## File check
        missing_files = utils.check_files(data_path)
        if missing_files:
            if download_files:
                utils.download_files(data_path, only_missing=only_missing)
            else:
                raise ValueError('There are missing data files. Files would have been downloaded if download_files=True (but it is not). Othersize if this was a mistake, please check/change the data_path: {}'.format(', '.join([os.path.split(f)[-1] for f in missing_files])))

        self.data_path = pathlib.Path(data_path)

        ## Open files for reading
        self._node = booklet.open(self.data_path.joinpath('node.blt'))
        self._catch = booklet.open(self.data_path.joinpath('catch.blt'))
        self._node_way_index = booklet.open(self.data_path.joinpath('node_way_index.blt'))
        self._way = booklet.open(self.data_path.joinpath('way.blt'))
        self._way_index = booklet.open(self.data_path.joinpath('way_index.blt'))
        self._way_tag = booklet.open(self.data_path.joinpath('way_tag.blt'))

        with open(self.data_path.joinpath('btree.node_ids.np.zstd'), 'rb') as f:
            self._all_node_ids = np.frombuffer(zstd.decompress(f.read()), 'i4')

        self._all_way_ids = tuple(self._way.keys())

        self.ids = None
        self._type = None

        ## Create spatial index
        with open(self.data_path.joinpath('btree.node_coords.np.zstd'), 'rb') as f:
            self._kdtree = KDTree(np.frombuffer(zstd.decompress(f.read()), 'i4').reshape(len(self._node), 2))


    def close(self):
        self._node.close()
        self._catch.close()
        self._node_way_index.close()
        self._way.close()
        self._way_index.close()
        self._way_tag.close()


    def __exit__(self, *args):
        self.close()


    def nearest_node(self, coords, max_distance=np.inf):
        """

        """
        dist, node, way = utils.nearest(self._kdtree, coords, self._all_node_ids, self._node_way_index, self._way, max_distance=max_distance)

        self.source_coords = coords
        self.distance = dist
        self._node_id = node
        self._way_id = way

        return Node(self)


    def nearest_way(self, coords, max_distance=np.inf):
        """

        """
        dist, node, way = utils.nearest(self._kdtree, coords, self._all_node_ids, self._node_way_index, self._way, max_distance=max_distance)

        self.source_coords = coords
        self.distance = dist
        self._node_id = node
        self._way_id = way

        return Way(self)


    def nearest_node_within_catchments(self, coords, max_distance=np.inf):
        """

        """
        dist, node, way = utils.nearest_within_catchments(self._kdtree, coords, self._all_node_ids, self._node, self._node_way_index, self._way, self._catch, max_distance=max_distance)

        self.source_coords = coords
        self.distance = dist
        self._node_id = node
        self._way_id = way

        return Node(self)


    def nearest_way_within_catchments(self, coords, max_distance=np.inf):
        """

        """
        dist, node, way = utils.nearest_within_catchments(self._kdtree, coords, self._all_node_ids, self._node, self._node_way_index, self._way, self._catch, max_distance=max_distance)

        self.source_coords = coords
        self.distance = dist
        self._node_id = node
        self._way_id = way

        return Way(self)


    def add_node(self, node_id=None):
        """

        """
        if node_id is None:
            if self._node_id is None:
                raise ValueError('node_id must be passed to the add_node method.')
        else:
            if node_id in self._all_node_ids:
                way_down = utils.get_down_way(node_id, self._node_way_index, self._way)
                self._node_id = node_id
                self._way_id = way_down
                self.distance = 0
                self.source_coords = None
            else:
                raise KeyError(node_id)

        return Node(self)


    def add_way(self, way_id=None):
        """

        """
        if way_id is None:
            if self._way_id is None:
                raise ValueError('way_id must be passed to the add_way method.')
        else:
            if way_id in self._all_way_ids:
                node_id = self._way[way_id][0]
                self._node_id = node_id
                self._way_id = way_id
                self.distance = 0
                self.source_coords = None
            else:
                raise KeyError(way_id)

        return Way(self)



class Feature:
    """

    """
    def coords(self):
        """

        """
        if issubclass(type(self), Node):
            node_coords = {int(i): self._water._node[int(i)] * 0.0000001 for i in self.nodes}
            return node_coords
        elif issubclass(type(self), Way):
            way_coords = {}
            for way_id in self.ways:
                nodes = self._water._way[int(way_id)]
                way_coords[int(way_id)] = np.array([self._water._node[int(i)] * 0.0000001 for i in nodes])
            return way_coords
        elif issubclass(type(self), Catchment):
            catch_coords = {i: self._water._catch[i] for i in self.ways}
            return catch_coords


    def to_gpd(self, add_attrs=False):
        """

        """
        if issubclass(type(self), Node):
            geo = [Point(self._water._node[int(i)] * 0.0000001) for i in self.nodes]
            gdf = gpd.GeoDataFrame({'node_id': self.nodes}, geometry=geo, crs=4326)
        elif issubclass(type(self), Way):
            geo = []
            for way_id in self.ways:
                nodes = self._water._way[int(way_id)]
                geo.append(LineString(np.array([self._water._node[int(i)] * 0.0000001 for i in nodes])))

            if add_attrs:
                data = []
                append = data.append
                for i in self.ways:
                    d1 = {'way_id': int(i)}
                    d1.update(self._water._way_tag[int(i)])
                    append(d1)
            else:
                data = [{'way_id': int(i)} for i in self.ways]

            gdf = gpd.GeoDataFrame(data, geometry=geo, crs=4326)

        elif issubclass(type(self), Catchment):
            geo = shapely.ops.unary_union([self._water._catch[int(i)] for i in self.ways]).buffer(0.0001)

            gdf = gpd.GeoDataFrame({'way_id': [self.id]}, geometry=[geo], crs=4326)

        return gdf


    def __repr__(self):
        if issubclass(type(self), Node):
            str1 = 'Node ID: {}'
        elif issubclass(type(self), (Way, Catchment)):
            str1 = 'Way ID: {}'
        return str1.format(self.id)


    def __enter__(self):
        return self






class Node(Feature):
    """

    """
    def __init__(self, water):
        """

        """
        self.id = water._node_id
        self._water = water
        self.nodes = set([water._node_id])
        self.distance = water.distance


    def ways(self):
        """

        """
        return Way(self._water)





class Way(Feature):
    """

    """
    def __init__(self, water):
        """

        """
        self.id = water._way_id
        self._water = water
        self.ways = set([water._way_id])
        self.distance = water.distance


    def nodes(self):
        """

        """
        return Node(self._water)


    def upstream(self):
        """

        """
        ways_up = utils.find_upstream(self._water._node_id, self._water._node_way_index, self._water._way, self._water._way_index)

        new_way = Way(self._water)
        new_way.ways.update(ways_up)

        return new_way


    def downstream(self):
        """

        """
        ways_down = utils.find_downstream(self._water._node_id, self._water._node_way_index, self._water._way, self._water._way_index)

        new_way = Way(self._water)
        new_way.ways.update(ways_down)

        return new_way


    def catchments(self):
        """

        """
        return Catchment(self)


    def attrs(self):
        """

        """
        data = []
        append = data.append
        for i in self.ways:
            d1 = {'way_id': int(i)}
            d1.update(self._water._way_tag[int(i)])
            append(d1)

        df1 = pd.DataFrame(data)

        return df1



class Catchment(Feature):
    """

    """
    def __init__(self, way_obj):
        """

        """
        self._water = way_obj._water
        self.id = way_obj.id
        self.ways = way_obj.ways




#####################################################
### Combo functions



def between(objs):
    """

    """
    ## Checks
    bool_way = all([isinstance(obj, Way) for obj in objs])
    bool_catch = all([isinstance(obj, Catchment) for obj in objs])

    new_list = []
    if bool_way or bool_catch:

        for obj in objs:
            diff_options = {len(obj.ways.difference(o.ways)): i for i, o in enumerate(objs) if obj.id != o.id}
            min_diff_len = min(diff_options)
            if min_diff_len > 0:
                next_up_way_pos = diff_options[min_diff_len]
                diff_ways = obj.ways.difference(objs[next_up_way_pos].ways)
                new_way_obj = Way(obj._water)
                new_way_obj.ways = diff_ways
            else:
                new_way_obj = obj

            if bool_catch:
                new_list.append(new_way_obj.catchments())
            else:
                new_list.append(new_way_obj)
    else:
        raise TypeError('Input objs must be all either Way or Cathcment objects.')

    return new_list






























































#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Dec 31 20:00:57 2022

@author: mike
"""
import io
import os
import geopandas as gpd
import pandas as pd
import numpy as np
import pathlib
from shapely.geometry import Point, Polygon, box, LineString
from copy import copy
from time import time
import booklet
import shapely
from scipy.spatial import KDTree
import zstandard as zstd
from hashlib import blake2b, blake2s

import utils

pd.options.display.max_columns = 10

##################################################
### Parameters

base_path = pathlib.Path(os.path.join(os.path.split(os.path.realpath(os.path.dirname(__file__)))[0], 'data'))

rec_rivers_clean = '/media/nvme1/data/NIWA/REC25_rivers/rec25_rivers_clean.feather'

rec_catch_clean = '/media/nvme1/data/NIWA/REC25_watersheds/rec25_watersheds_clean.feather'

node_blt = 'node.blt'
way_blt = 'way.blt'
way_tag_blt = 'way_tag.blt'
node_way_blt = 'node_way_index.blt'
way_index_blt = 'way_index.blt'
catch_blt = 'catch.blt'


# class NodeSerial:
#     def dumps(obj):
#         return (obj.round(7) * 10000000).astype('i4').tobytes()
#     def loads(obj):
#         return np.frombuffer(obj, 'i4') * 0.0000001

tag_files = {
    'NZRiverMaps_bed_sediment_2023-01-14.csv': ('nzsegment', 'Catchment name', 'Strahler stream order', 'Region', 'Bedrock', 'Boulder', 'CoarseGravel', 'Cobble', 'FineGravel', 'Mud', 'Sand'),
    'NZRiverMaps_bird_habitat_2023-01-14.csv': ('Whio (blue duck)'),
    'NZRiverMaps_fish_habitat_2023-01-14.csv': ('Koaro total WUA low flow', 'Koaro WUA_MALF / WW_Median', 'Longfin total WUA low flow', 'Longfin WUA_MALF / WW_Median', 'Shortfin total WUA low flow', 'Shortfin WUA_MALF / WW_Median', 'Smelt total WUA low flow', 'Smelt WUA_MALF / WW_Median', 'Torrentfish total WUA low flow', 'Torrentfish WUA_MALF / WW_Median', 'Trout total WUA low flow', 'Trout WUA_MALF / WW_Median'),
    'NZRiverMaps_fish_presence_2023-01-14.csv': ("Alpine galaxias","Banded kokopu","Bignose galaxias","Black flounder","Bluegill bully","Brook char","Brook char (Potential)","Brown trout","Brown trout (Potential)","Canterbury galaxias","Chinook salmon","Clutha flathead galaxias","Common bully","Common smelt","Crans bully","Dwarf galaxias","Flathead galaxias","Gambusia","Gambusia (Potential)","Giant bully","Giant kokopu","Goldfish","Goldfish (Potential)","Gollum galaxias","Inanga","Koaro","Lamprey","Longfin eel","Northern flathead galaxias","Rainbow trout","Rainbow trout (Potential)","Redfin bully","Roundhead galaxias","Shortfin eel","Shortjaw kokopu","Torrentfish","Upland bully","Upland longjaw galaxias"),
    'NZRiverMaps_hydrology_2023-01-09.csv': ("1 in 5 year low flow","February flow seasonality","FRE3","MALF","Mean Flow","Median flow","Month lowest mean flow"),
    'NZRiverMaps_invertebrate_2023-01-14.csv': ("% EPT Richness","EPT","EPTNoHydrop","Invertebrate Taxa Richness","MCI (2015)","MCI (2019)","MCI (2021)","Number of EPT taxa","Number of taxa","SQMCI-hb"),
    'NZRiverMaps_isotopes_2023-01-14.csv': ("delta 18O","delta 2H"),
    'NZRiverMaps_rec_2023-01-09.csv': ("Climate class","Geology class","Landcover class","Network position class","Topography class","Upstream catchment area","Valley landform class"),
    'NZRiverMaps_sediment_load_2023-01-14.csv': ("Suspended sediment load","Suspended sediment load lake percent change ","Suspended sediment load no lake adjustment"),
    'NZRiverMaps_wetted_width_2023-01-14.csv': ("Width at MALF","Width at mean flow","Width at median flow","Width at Q5"),
    'NZRiverMaps_WQ_2023-01-14.csv': ("Ammoniacal nitrogen AnnMax pH","Ammoniacal nitrogen median","Ammoniacal nitrogen median pH","ANZECC elevation class","CHLA 92%","CHLA Mean","Dissolved oxygen","Dissolved oxygen saturation","Dissolved reactive phosphorus median","Dissolved reactive phosphorus Q95","E. coli G260","E. coli G540","E. coli median","E. coli Q95","Nitrate + nitrite median","Nitrate + nitrite Q95","Temperature","Total nitrogen median","Total phosphorus median","Total suspended solids","Turbidity median","Visual clarity median","WCC 92%","WCC Mean"),
    }

tag_float_cols = ['Bedrock', 'Boulder', 'CoarseGravel', 'Cobble', 'FineGravel', 'Mud', 'Sand', 'Whio (blue duck)', 'Koaro total WUA low flow', 'Koaro WUA_MALF / WW_Median', 'Longfin total WUA low flow', 'Longfin WUA_MALF / WW_Median', 'Shortfin total WUA low flow', 'Shortfin WUA_MALF / WW_Median', 'Smelt total WUA low flow', 'Smelt WUA_MALF / WW_Median', 'Torrentfish total WUA low flow', 'Torrentfish WUA_MALF / WW_Median', 'Trout total WUA low flow', 'Trout WUA_MALF / WW_Median', "1 in 5 year low flow", "February flow seasonality", "FRE3","MALF","Mean Flow","Median flow", "% EPT Richness","EPT","EPTNoHydrop","Invertebrate Taxa Richness","MCI (2015)","MCI (2019)","MCI (2021)","Number of EPT taxa","Number of taxa","SQMCI-hb", "delta 18O","delta 2H", "Upstream catchment area", "Suspended sediment load","Suspended sediment load lake percent change ","Suspended sediment load no lake adjustment", "Width at MALF","Width at mean flow","Width at median flow","Width at Q5", "Ammoniacal nitrogen AnnMax pH","Ammoniacal nitrogen median","Ammoniacal nitrogen median pH","ANZECC elevation class","CHLA 92%","CHLA Mean","Dissolved oxygen","Dissolved oxygen saturation","Dissolved reactive phosphorus median","Dissolved reactive phosphorus Q95","E. coli G260","E. coli G540","E. coli median","E. coli Q95","Nitrate + nitrite median","Nitrate + nitrite Q95","Temperature","Total nitrogen median","Total phosphorus median","Total suspended solids","Turbidity median","Visual clarity median","WCC 92%","WCC Mean"]

float_precision = 4

###############################################
### Rivers

rivers0 = gpd.read_feather(rec_rivers_clean)
rivers1 = rivers0.to_crs(4326)
# rivers1['nzsegment'] = rivers1.nzsegment.astype('int32')
# rivers1['from_node'] = rivers1.from_node.astype('int32')
# rivers1['to_node'] = rivers1.to_node.astype('int32')

### Make the dicts
# dict_start = time()
way_dict = {}
node_dict = {}
way_tag_dict = {}

node_counter = 1000000
for i, row in rivers1.iterrows():
    # print(i)
    # Nodes
    tmp_node = {}
    nodes = list(row.geometry.coords)
    tmp_node[row.from_node] = np.array(nodes[0])

    for n in nodes[1:-1]:
        node_id = node_counter
        node_counter += 1
        tmp_node[node_id] = np.array(n)

    tmp_node[row.to_node] = np.array(nodes[-1])

    node_dict.update(tmp_node)

    # Ways
    way_id = int(row.nzsegment)
    way_dict[way_id] = list(tmp_node.keys())
    way_tag_dict[way_id] = {'Strahler stream order': row.stream_order}

# dict_end = time()

# print((dict_end - dict_start)/60)

## node to way index
node_way_dict = {}
for i, wa in way_dict.items():
    # print(i)
    for n in wa:
        if n in node_way_dict:
            node_way_dict[int(n)].append(i)
        else:
            node_way_dict[int(n)] = [i]

## way to way index
way_index_dict = {}
for w, nodes in way_dict.items():
    waynode = []
    for node in nodes:
        if node in node_way_dict:
            ways = copy(node_way_dict[node])
            ways.remove(w)
            if ways:
                waynode.extend(ways)
    way_index_dict[w] = waynode

d

### Make the special persistant dicts
with booklet.open(base_path.joinpath(way_blt), 'n', value_serializer='numpy_int4', key_serializer='uint4', n_bytes_value=2, n_buckets=100000) as way:
    for k, v in way_dict.items():
        way[k] = np.array(v)

with booklet.open(base_path.joinpath(node_blt), 'n', value_serializer='numpy_int4', key_serializer='uint4', n_bytes_value=1, n_buckets=500000) as node:
    for k, v in node_dict.items():
        node[k] = (v.round(7) * 10000000).astype('i4')

# with booklet.open(base_path.joinpath(way_tag_blt), 'n', value_serializer='orjson_zstd', key_serializer='uint4', n_bytes_value=1, n_buckets=100000) as way_tag:
#     for k, v in way_tag_dict.items():
#         way_tag[k] = v

with booklet.open(base_path.joinpath(node_way_blt), 'n', value_serializer='numpy_int4', key_serializer='uint4', n_bytes_value=1, n_buckets=500000) as node_way:
    for k, v in node_way_dict.items():
        node_way[k] = np.array(v)

with booklet.open(base_path.joinpath(way_index_blt), 'n', value_serializer='numpy_int4', key_serializer='uint4', n_bytes_value=1, n_buckets=100000) as way_index:
    for k, v in way_index_dict.items():
        way_index[k] = np.array(v)


### Catchments
catch0 = gpd.read_feather(rec_catch_clean)
catch1 = catch0.to_crs(4326)

with booklet.open(base_path.joinpath(catch_blt), 'n', value_serializer='wkb_zstd', key_serializer='uint4', n_bytes_value=3, n_buckets=100000) as catch:
    for k, v in catch1.iterrows():
        catch[v.nzsegment] = v.geometry


## Btree index
node = booklet.open(base_path.joinpath(node_blt))
node_ids = np.array(list(node.keys()), dtype='int32')

nodes_coords = np.array([v for v in node.values()])
# node_ids = np.array([v for v in node.keys()], dtype='i4')

# btree = KDTree(nodes_coords, copy_data=True)

with open('/media/nvme1/git/nzrec/data/btree.node_coords.np.zstd', 'wb') as f:
    f.write(zstd.compress(nodes_coords.tobytes(), 20))

# with open('/media/nvme1/git/nzrec/data/btree.node_coords.np.zstd', 'rb') as f:
#     nodes3 = np.frombuffer(zstd.decompress(f.read()), 'i4').reshape(nodes_coords.shape)

with open('/media/nvme1/git/nzrec/data/btree.node_ids.np.zstd', 'wb') as f:
    f.write(zstd.compress(node_ids.tobytes(), 20))

### Way tags
rec_rivers_base_path = pathlib.Path(rec_rivers_clean).parent
tag_df_list = []
for file, cols in tag_files.items():
    print(file)
    if isinstance(cols, str):
        cols = [cols]
    df1 = pd.read_csv(rec_rivers_base_path.joinpath(file), usecols=list(cols))
    tag_df_list.append(df1)

combo_tag = pd.concat(tag_df_list, axis=1)
combo_tag = combo_tag.set_index('nzsegment')

combo_tag = combo_tag[combo_tag.index.isin(set(way_tag_dict.keys()))].copy()

missing_segs = list(set(way_tag_dict.keys()).difference(set(combo_tag.index)))

stream_orders = []
for seg_id in missing_segs:
    stream_orders.append(way_tag_dict[seg_id]['Strahler stream order'])

mis_seg_df = pd.Series(stream_orders, index=missing_segs)
mis_seg_df.name = 'Strahler stream order'

combo_tag = pd.concat([combo_tag, mis_seg_df.to_frame()])

combo_tag.index = combo_tag.index.astype('int32')
combo_tag['Strahler stream order'] = combo_tag['Strahler stream order'].astype(int)
combo_tag[tag_float_cols] = combo_tag[tag_float_cols].round(float_precision)


with booklet.open(base_path.joinpath(way_tag_blt), 'n', value_serializer='orjson_zstd', key_serializer='uint4', n_bytes_value=3, n_buckets=100000) as way_tag:
    for seg_id, row in combo_tag.iterrows():
        way_tag[seg_id] = row.to_dict()














############################################################
### Testing




# node_id = 590718
# node_id = 436822

# node_way = booklet.open(base_path.joinpath(node_way_blt))
# way = booklet.open(base_path.joinpath(way_blt))
# way_index = booklet.open(base_path.joinpath(way_index_blt))
# node = booklet.open(base_path.joinpath(node_blt))
# catch = booklet.open(base_path.joinpath(catch_blt))

# ways_up = find_upstream(node_id, node_way, way, way_index)

# reach_gpd = make_reaches(ways_up, way, node)

# reach_gpd.to_file(base_path.joinpath('test_reach.gpkg'))


# way_down = find_downstream(node_id, node_way, way, way_index)

# catch_gpd = gpd.GeoDataFrame([node_id], geometry=[catch1], columns=['way_id'], crs=4326)

# catch_gpd.to_file(base_path.joinpath('test_catch.gpkg'))






# keys = list(node_dict.keys())

# mod_set = set()
# mod_dict = {}
# for key in keys:
#     k_int = int.from_bytes(blake2b(key.to_bytes(4, 'little'), digest_size=13).digest(), 'little')
#     mod = k_int % 9765625
#     if mod in mod_set:
#         if mod in mod_dict:
#             mod_dict[mod] += mod_dict[mod]
#         else:
#             mod_dict[mod] = 2
#     mod_set.add(mod)

























#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct 26 16:07:35 2022

@author: mike
"""
from tethysts import Tethys
import os

#############################################
### Parameters

base_path = os.path.join(os.path.split(os.path.realpath(os.path.dirname(__file__)))[0], 'datasets')

remotes = [{'bucket': 'nz-open-modelling-consortium', 'public_url': 'https://b2.nzrivers.xyz/file/', 'version': 4},
           {'bucket': 'fire-emergency-nz', 'public_url': 'https://b2.tethys-ts.xyz/file/', 'version': 4},
           {'bucket': 'ecan-env-monitoring', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4},
           {'bucket': 'gwrc-env', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
           ]

dataset_ids = {'7751c5f1bf47867fb109d7eb': [10], '0b2bd62cc42f3096136f11e9': None, 'f16774ea29f024a306c7fc7a': None, '9568f663d566aabb62a8e98e': None}

############################################
### Get data

t1 = Tethys(remotes)

for ds_id, heights in dataset_ids.items():
    stns1 = t1.get_stations(ds_id)
    station_ids = [s['station_id'] for s in stns1[:2]]
    for stn_id in station_ids:
        file_name = '{ds_id}_{stn_id}.nc'.format(ds_id=ds_id, stn_id=stn_id)
        file_path = os.path.join(base_path, file_name)
        results = t1.get_results(ds_id, stn_id, heights=heights)
        results.to_netcdf(file_path, engine='h5netcdf')



















































































#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Nov 29 20:02:57 2022

@author: mike
"""
import io
import os
import geopandas as gpd
import pandas as pd
import numpy as np
import utils
from shapely.geometry import Point, Polygon, box, LineString

pd.options.display.max_columns = 10

##################################################
### Parameters

base_path = os.path.join(os.path.split(os.path.realpath(os.path.dirname(__file__)))[0], 'data')

rec_rivers_shp = '/media/nvme1/data/NIWA/REC25_rivers/rec25_rivers.shp'
rec_catch_shp ='/media/nvme1/data/NIWA/REC25_watersheds/rec25_watersheds.shp'

segment_id_col = 'nzsegment'

# rec_rivers_clean_gpkg = '/media/nvme1/data/NIWA/REC25_rivers/rec25_rivers_clean.gpkg'

rec_rivers_clean_shp = '/media/nvme1/data/NIWA/REC25_rivers/rec25_rivers_clean.shp'
rec_rivers_clean_feather = '/media/nvme1/data/NIWA/REC25_rivers/rec25_rivers_clean.feather'

rec_catch_clean_shp = '/media/nvme1/data/NIWA/REC25_watersheds/rec25_watersheds_clean.shp'
rec_catch_clean_feather = '/media/nvme1/data/NIWA/REC25_watersheds/rec25_watersheds_clean.feather'

#################################################
### Rivers

rec_rivers0 = gpd.read_file(rec_rivers_shp)

## Select basic columns and simplify slightly
rec_rivers1 = rec_rivers0[['nzsegment', 'FROM_NODE', 'TO_NODE', 'geometry']].rename(columns={'FROM_NODE': 'from_node', 'TO_NODE': 'to_node'}).copy()
rec_rivers1['from_node'] = rec_rivers1.from_node.astype('int32')
rec_rivers1['to_node'] = rec_rivers1.to_node.astype('int32')
rec_rivers1['nzsegment'] = rec_rivers1.nzsegment.astype('int32')

rec_rivers1['geometry'] = rec_rivers1['geometry'].simplify(1)

## Find all end segments
end_segs = []
append = end_segs.append
for i, seg in rec_rivers1[['nzsegment', 'from_node', 'to_node']].iterrows():
    # print(i)
    seg_bool = rec_rivers1.from_node == seg.to_node
    if not seg_bool.any():
        append(seg.nzsegment)

## Make sure all lines are oriented downstream and remove lone 1st order streams
rec_rivers1.set_index('nzsegment', inplace=True)

lone_end_seg_list = []
end_seg_list = []
for seg in end_segs:
    row = rec_rivers1.loc[seg].copy()

    # Determine end seg orientation
    other1 = rec_rivers1[rec_rivers1.to_node == row.from_node]

    if not other1.empty:
        row2 = other1.iloc[0]
        pts0 = utils.convert_line_to_points(row.geometry, 0, 'array')
        pts2 = utils.convert_line_to_points(row2.geometry, 0, 'array')

        if (pts0[0] == pts2[-1]).all() or (pts0[0] == pts2[0]).all():
            pass
        elif (pts0[-1] == pts2[-1]).all() or (pts0[-1] == pts2[0]).all():
            print('reverse')
            pts0.reverse()
            row['geometry'] = LineString(pts0)
        else:
            raise ValueError('What is going on...')

        dict1 = row.to_dict()
        dict1['nzsegment'] = row.name
        end_seg_list.append(dict1)
    else:
        lone_end_seg_list.append(seg)
        # print(seg)
        # print('lonely segment')


## Are the lone ones at the upper end just reversed?
from_from_list = []
from_to_list = []

for seg in lone_end_seg_list:
    row = rec_rivers1.loc[seg].copy()

    # Determine end seg orientation
    other1 = rec_rivers1[rec_rivers1.from_node == row.from_node]
    other2 = rec_rivers1[rec_rivers1.from_node == row.to_node]

    if len(other1) > 1:
        from_from_list.append(seg)
    if not other2.empty:
        from_to_list.append(seg)


## Test all upstream segments and correct direction
reaches_lst = []
append = reaches_lst.append
for seg in end_segs:
    print(seg)
    reach1 = rec_rivers1.loc[[seg]].copy()
    append(reach1)
    up1 = rec_rivers1[rec_rivers1.to_node.isin(reach1.from_node)]
    while not up1.empty:
        up_list = []
        for i, row in up1.iterrows():
            base = reach1[reach1.from_node == row.to_node].iloc[0]
            pts0 = utils.convert_line_to_points(base.geometry, 0, 'array')
            pts2 = utils.convert_line_to_points(row.geometry, 0, 'array')

            if (pts0[0] == pts2[-1]).all():
                pass
            else:
                print('reverse')
                print(row)
                pts2 = utils.convert_line_to_points(row.geometry, 3, 'array')
                pts2.reverse()
                row['geometry'] = LineString(pts2)

            row_df = row.to_frame().T
            append(row_df)
            up_list.append(row_df)

        reach1 = pd.concat(up_list)
        up1 = rec_rivers1[rec_rivers1.to_node.isin(reach1.from_node)]

for seg in lone_end_seg_list:
    reach1 = rec_rivers1.loc[[seg]].copy()
    reaches_lst.append(reach1)

reaches = pd.concat(reaches_lst)
reaches.index.name = 'nzsegment'
reaches = reaches.reset_index().drop_duplicates(subset='nzsegment')

# Save cleaned up data
reaches = reaches.merge(rec_rivers0[['nzsegment', 'StreamOrde']].rename(columns={'StreamOrde': 'stream_order'}), on='nzsegment')
reaches.crs = rec_rivers0.crs
# reaches['geometry'] = reaches['geometry'].simplify(0)

reaches.to_file(rec_rivers_clean_shp)

# reaches = gpd.read_file(rec_rivers_clean_shp)
# reaches = reaches.rename(columns={'stream_ord': 'stream_order'})
reaches['nzsegment'] = reaches['nzsegment'].astype('int32')
reaches['from_node'] = reaches['from_node'].astype('int32')
reaches['to_node'] = reaches['to_node'].astype('int32')
reaches['stream_order'] = reaches['stream_order'].astype('int8')

reaches.to_feather(rec_rivers_clean_feather, compression='zstd')

# b1 = io.BytesIO()

# reaches.to_feather(b1, compression='zstd')


del rec_rivers0
del rec_rivers1
del reaches

######################################################
### Watersheds

rec_catch0 = gpd.read_file(rec_catch_shp)

rec_catch1 = rec_catch0[['nzsegment', 'geometry']].copy()

rec_catch1['geometry'] = rec_catch1.buffer(0)
rec_catch1['nzsegment'] = rec_catch1['nzsegment'].astype('int32')

rec_catch1.to_file(rec_catch_clean_shp)

# rec_catch1 = gpd.read_file(rec_catch_clean_shp)

rec_catch1.to_feather(rec_catch_clean_feather, compression='zstd')


































































